@@ -3,6 +3,7 @@ import tempfile
 from collections import namedtuple
 
 import numpy
+import torch.jit
 
 from einops import rearrange, reduce
 from einops.einops import _reductions
@@ -79,16 +80,10 @@ def test_rearrange_symbolic():
                 result1 = backend.eval_symbol(result_symbol1, eval_inputs)
                 assert numpy.allclose(result_numpy, result1)
 
-                if 'keras' not in backend.framework_name:
-                    # simple pickling / unpickling
-                    # keras bug - fails for pickling
-                    layer2 = pickle.loads(pickle.dumps(layer))
-                    result_symbol2 = layer2(symbol)
-                    result2 = backend.eval_symbol(result_symbol2, eval_inputs)
-                    assert numpy.allclose(result1, result2)
-                else:
-                    # keras is deprecated, so no checks for save/load
-                    pass
+                layer2 = pickle.loads(pickle.dumps(layer))
+                result_symbol2 = layer2(symbol)
+                result2 = backend.eval_symbol(result_symbol2, eval_inputs)
+                assert numpy.allclose(result1, result2)
 
                 # now testing back-propagation
                 just_sum = backend.layers().Reduce('...->', reduction='sum')
@@ -176,39 +171,40 @@ def test_reduce_symbolic():
                     result1 = backend.eval_symbol(result_symbol1, eval_inputs)
                     assert numpy.allclose(result_numpy, result1)
 
-                    if 'keras' not in backend.framework_name:
-                        # simple pickling / unpickling
-                        # keras bug - fails for pickling, requires special import/export
-                        layer2 = pickle.loads(pickle.dumps(layer))
-                        result_symbol2 = layer2(symbol)
-                        result2 = backend.eval_symbol(result_symbol2, eval_inputs)
-                        assert numpy.allclose(result1, result2)
-                    else:
-                        pass
+                    layer2 = pickle.loads(pickle.dumps(layer))
+                    result_symbol2 = layer2(symbol)
+                    result2 = backend.eval_symbol(result_symbol2, eval_inputs)
+                    assert numpy.allclose(result1, result2)
+
+
+def create_torch_model(use_reduce=False, add_scripted_layer=False):
+    from torch.nn import Sequential, Conv2d, MaxPool2d, Linear, ReLU
+    from einops.layers.torch import Rearrange, Reduce, EinMix
+    return Sequential(
+        Conv2d(3, 6, kernel_size=(5, 5)),
+        Reduce('b c (h h2) (w w2) -> b c h w', 'max', h2=2, w2=2) if use_reduce else MaxPool2d(kernel_size=2),
+        Conv2d(6, 16, kernel_size=(5, 5)),
+        Reduce('b c (h h2) (w w2) -> b c h w', 'max', h2=2, w2=2),
+        torch.jit.script(Rearrange('b c h w -> b (c h w)'))
+        if add_scripted_layer else Rearrange('b c h w -> b (c h w)'),
+        Linear(16 * 5 * 5, 120),
+        ReLU(),
+        Linear(120, 84),
+        ReLU(),
+        EinMix('b c1 -> (b c2)', weight_shape='c1 c2', bias_shape='c2', c1=84, c2=84),
+        EinMix('(b c2) -> b c3', weight_shape='c2 c3', bias_shape='c3', c2=84, c3=84),
+        Linear(84, 10),
+    )
 
 
 def test_torch_layer():
-    if any(backend.framework_name == 'torch' for backend in collect_test_backends(symbolic=False, layers=True)):
+    has_torch = any(backend.framework_name == 'torch' for backend in collect_test_backends(symbolic=False, layers=True))
+    if has_torch:
         # checked that torch present
         import torch
-        from torch.nn import Sequential, Conv2d, MaxPool2d, Linear, ReLU
-        from einops.layers.torch import Rearrange, Reduce
 
-        def create_model(use_reduce=False):
-            return Sequential(
-                Conv2d(3, 6, kernel_size=5),
-                Reduce('b c (h h2) (w w2) -> b c h w', 'max', h2=2, w2=2) if use_reduce else MaxPool2d(kernel_size=2),
-                Conv2d(6, 16, kernel_size=5),
-                Reduce('b c (h h2) (w w2) -> b (c h w)', 'max', h2=2, w2=2),
-                Linear(16 * 5 * 5, 120),
-                ReLU(),
-                Linear(120, 84),
-                ReLU(),
-                Linear(84, 10),
-            )
-
-        model1 = create_model(use_reduce=True)
-        model2 = create_model(use_reduce=False)
+        model1 = create_torch_model(use_reduce=True)
+        model2 = create_torch_model(use_reduce=False)
         input = torch.randn([10, 3, 32, 32])
         # random models have different predictions
         assert not torch.allclose(model1(input), model2(input))
@@ -225,16 +221,26 @@ def test_torch_layer():
         torch.testing.assert_allclose(model1(input + 1), model4(input + 1), atol=1e-3, rtol=1e-3)
 
 
+def test_torch_layers_scripting():
+    import torch
+    for script_layer in [False, True]:
+        model1 = create_torch_model(use_reduce=True, add_scripted_layer=script_layer)
+        model2 = torch.jit.script(model1)
+        input = torch.randn([10, 3, 32, 32])
+
+        torch.testing.assert_allclose(model1(input), model2(input), atol=1e-3, rtol=1e-3)
+
+
 def test_keras_layer():
-    if any(backend.framework_name == 'keras' for backend in collect_test_backends(symbolic=True, layers=True)):
+    if any(backend.framework_name == 'tensorflow.keras' for backend in collect_test_backends(symbolic=True, layers=True)):
         # checked that keras present
 
         import tensorflow as tf
         from tensorflow.keras.models import Sequential
-        from tensorflow.keras.layers import MaxPool2D as MaxPool2d, Conv2D as Conv2d, Dense as Linear, ReLU
-        from einops.layers.keras import Rearrange, Reduce, keras_custom_objects
+        from tensorflow.keras.layers import Conv2D as Conv2d, Dense as Linear, ReLU
+        from einops.layers.keras import Rearrange, Reduce, EinMix, keras_custom_objects
 
-        def create_model():
+        def create_keras_model():
             return Sequential([
                 Conv2d(6, kernel_size=5, input_shape=[32, 32, 3]),
                 Reduce('b c (h h2) (w w2) -> b c h w', 'max', h2=2, w2=2),
@@ -245,11 +251,13 @@ def test_keras_layer():
                 ReLU(),
                 Linear(84),
                 ReLU(),
+                EinMix('b c1 -> (b c2)', weight_shape='c1 c2', bias_shape='c2', c1=84, c2=84),
+                EinMix('(b c2) -> b c3', weight_shape='c2 c3', bias_shape='c3', c2=84, c3=84),
                 Linear(10),
             ])
 
-        model1 = create_model()
-        model2 = create_model()
+        model1 = create_keras_model()
+        model2 = create_keras_model()
         input = numpy.random.normal(size=[10, 32, 32, 3]).astype('float32')
         assert not numpy.allclose(model1.predict_on_batch(input), model2.predict_on_batch(input))
 
@@ -272,11 +280,14 @@ def test_keras_layer():
 
 
 def test_gluon_layer():
-    if any('mxnet' in backend.framework_name for backend in collect_test_backends(symbolic=False, layers=True)):
+    gluon_is_present = any(
+        'mxnet' in backend.framework_name for backend in collect_test_backends(symbolic=False, layers=True)
+    )
+    if gluon_is_present:
         # checked that gluon present
         import mxnet
         from mxnet.gluon.nn import HybridSequential, Dense, Conv2D, LeakyReLU
-        from einops.layers.gluon import Rearrange, Reduce
+        from einops.layers.gluon import Rearrange, Reduce, EinMix
         from einops import asnumpy
 
         def create_model():
@@ -323,8 +334,50 @@ def test_gluon_layer():
         assert numpy.allclose(asnumpy(model3(x)), asnumpy(model4(x)))
 
         try:
-            # hybridization doesn't work
             model1.hybridize(static_alloc=True, static_shape=True)
             model1(x)
         except:
+            # hybridization is not supported
             pass
+
+
+def test_chainer_layer():
+    chainer_is_present = any(
+        'chainer' in backend.framework_name for backend in collect_test_backends(symbolic=False, layers=True)
+    )
+    if chainer_is_present:
+        # checked that gluon present
+        import chainer
+        import chainer.links as L
+        import chainer.functions as F
+        from einops.layers.chainer import Rearrange, Reduce, EinMix
+        from einops import asnumpy
+        import numpy as np
+
+        def create_model():
+            return chainer.Sequential(
+                L.Convolution2D(3, 6, ksize=(5, 5)),
+                Reduce('b c (h h2) (w w2) -> b c h w', 'max', h2=2, w2=2),
+                L.Convolution2D(6, 16, ksize=(5, 5)),
+                Reduce('b c (h h2) (w w2) -> b c h w', 'max', h2=2, w2=2),
+                Rearrange('b c h w -> b (c h w)'),
+                L.Linear(16 * 5 * 5, 120),
+                L.Linear(120, 84),
+                F.relu,
+                EinMix('b c1 -> (b c2)', weight_shape='c1 c2', bias_shape='c2', c1=84, c2=84),
+                EinMix('(b c2) -> b c3', weight_shape='c2 c3', bias_shape='c3', c2=84, c3=84),
+                L.Linear(84, 10),
+            )
+
+        model1 = create_model()
+        model2 = create_model()
+        x = np.random.normal(size=[10, 3, 32, 32]).astype('float32')
+        x = chainer.Variable(x)
+        assert not numpy.allclose(asnumpy(model1(x)), asnumpy(model2(x)))
+
+        with tempfile.TemporaryDirectory() as dir:
+            filename = f'{dir}/file.npz'
+            chainer.serializers.save_npz(filename, model1)
+            chainer.serializers.load_npz(filename, model2)
+
+        assert numpy.allclose(asnumpy(model1(x)), asnumpy(model2(x)))
